
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/times.h>
#include <stdexcept>
#include <Magick++.h>

extern void _merge_with_alpha_c(unsigned char *d, const unsigned char *s1, const unsigned char *s2, int w, int h);
extern void _merge_with_alpha_mmx(unsigned char *d, const unsigned char *s1, const unsigned char *s2, int w, int h);


class image
{
public:
	image(const char *filename)
		: f_width(0)
		, f_height(0)
		, f_rgba(NULL)
	{
		Magick::Image image;
		image.read(filename);
		Magick::Geometry g(image.size());
		f_width = g.width();
		f_height = g.height();
		int e(posix_memalign(reinterpret_cast<void **>(&f_rgba), 16, f_width * f_height * 4));
		if(e != 0)
		{
			fprintf(stderr, "error: could not allocate image memory buffer. (errno: %d)\n", e);
			throw std::bad_alloc();
		}
		Magick::Blob blob;
		image.magick("RGBA");
		image.write(&blob);
		if(blob.length() != f_width * f_height * 4)
		{
			fprintf(stderr, "error: blob length() not equal width x height x depth.\n");
			throw std::runtime_error("incompatible sizes");
		}
		memcpy(f_rgba, blob.data(), f_width * f_height * 4);
	}

	image(int width, int height, bool clear = true)
		: f_width(width)
		, f_height(height)
		, f_rgba(NULL)
	{
		int e(posix_memalign(reinterpret_cast<void **>(&f_rgba), 16, f_width * f_height * 4));
		if(e != 0)
		{
			fprintf(stderr, "error: could not allocate image memory buffer. (errno: %d)\n", e);
			throw std::bad_alloc();
		}
		if(clear)
		{
			memset(f_rgba, 0, f_width * f_height * 4);
		}
	}
	~image()
	{
		free(f_rgba);
	}

	void save(const char *filename)
	{
		Magick::Image image(f_width, f_height, "RGBA", Magick::CharPixel, f_rgba);
		image.write(filename);
	}

	int f_width;
	int f_height;
	unsigned char *f_rgba;
};

class run_timer
{
public:
	run_timer()
	{
	}

	void start()
	{
		times(&f_start);
	}

	void stop()
	{
		times(&f_stop);
	}

	void report(const char *msg)
	{
		printf("%s:\n"
		       "Function started at %ld\n"
		       "           ended at %ld\n"
		       "              spent %ld (%f seconds)\n",
		       msg,
		       f_start.tms_utime,
		       f_stop.tms_utime,
		       f_stop.tms_utime - f_start.tms_utime,
		       (double)(f_stop.tms_utime - f_start.tms_utime)/(double)sysconf(_SC_CLK_TCK));
	}

	struct tms f_start;
	struct tms f_stop;
};


int main(int argc, char *argv[])
{
	run_timer t;

	image s1("foreground.png");
	image s2("background.png");

	image d(640, 480);
	t.start();
	for(int i(0); i < 3300; ++i)
	{
		_merge_with_alpha_c(d.f_rgba, s1.f_rgba, s2.f_rgba, 640, 480);
	}
	t.stop();
	t.report("C version");
	d.save("c.png");

	image e(640, 480);
	t.start();
	for(int i(0); i < 3300; ++i)
	{
		_merge_with_alpha_mmx(e.f_rgba, s1.f_rgba, s2.f_rgba, 640, 480);
	}
	t.stop();
	t.report("SIMD version");
	e.save("mmx.png");

	return 0;
}

